﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;

namespace ImageCollage {

	public partial class ImageCollageApp : Form {

		public ImageCollage Collage { get; private set; }

		public Bitmap Bitmap { get; private set; }

		private List<string> mainImages;
		private List<string> faveImages;
        private TileSize tileSizeMode;
        private bool imagesUnchanged;
        private bool previewOnly;
        private bool skipLayout;

		public ImageCollageApp() {
			Collage = new ImageCollage();
			Collage.Width = 1600f;
			Collage.Height = 1200f;
			Collage.BackColor = Color.Black;

			InitializeComponent();

			backgroundWorker.DoWork += backgroundWorker_DoWork;
			backgroundWorker.RunWorkerCompleted += backgroundWorker_RunWorkerCompleted;
			backgroundWorker.ProgressChanged += backgroundWorker_ProgressChanged;

			lstFavourites.DragEnter += lstFavourites_DragEnter;
			lstFavourites.DragDrop += lstFavourites_DragDrop;
			lstFavourites.MouseMove += lstFavourites_MouseMove;
			lstImages.DragEnter += lstImages_DragEnter;
			lstImages.DragDrop += lstImages_DragDrop;
			lstImages.MouseMove += lstImages_MouseMove;

			lstImages.Format += lstImages_Format;
			lstFavourites.Format += lstFavourites_Format;

			btnGo.Click += btnGo_Click;
			btnOpen.Click += btnOpen_Click;
			btnSave.Click += btnSave_Click;
            btnPreview.Click += btnPreview_Click;

			lstImages.KeyDown += lstImages_KeyDown;
            lstFavourites.KeyDown += lstFavourites_KeyDown;

            // menus
            addImagesToolStripMenuItem.Click += btnOpen_Click;
            clearImagesToolStripMenuItem.Click += clearImagesToolStripMenuItem_Click;
            saveCollageToolStripMenuItem.Click += btnSave_Click;
            exitToolStripMenuItem.Click += exitToolStripMenuItem_Click;

            previewLayoutToolStripMenuItem.Click += btnPreview_Click;
            generateCollageToolStripMenuItem.Click += btnGo_Click;

            aboutToolStripMenuItem.Click += aboutToolStripMenuItem_Click;

            // mutually exclusive menu options
            uniformToolStripMenuItem.Click += mutualExclusionHandler;
            randomToolStripMenuItem.Click += mutualExclusionHandler;

            blackToolStripMenuItem.Click += mutualExclusionHandler;
            whiteToolStripMenuItem.Click += mutualExclusionHandler;
            transparentToolStripMenuItem.Click += mutualExclusionHandler;

            small640x480ToolStripMenuItem.Click += mutualExclusionHandler;
            medium1024x768ToolStripMenuItem.Click += mutualExclusionHandler;
            large1600x1200ToolStripMenuItem.Click += mutualExclusionHandler;

            averagedToolStripMenuItem.Click += mutualExclusionHandler;
            preserveHorizontalToolStripMenuItem.Click += mutualExclusionHandler;
            preserveVerticalToolStripMenuItem.Click += mutualExclusionHandler;
		}

        private void Go(bool preview) {
            skipLayout = (previewOnly && !preview);
            previewOnly = preview;

            if (Bitmap != null) {
                picResult.Image = null;
                Bitmap.Dispose();
                Bitmap = null;
            }

            List<string> oldMain = mainImages;
            List<string> oldFave = faveImages;

            mainImages = new List<string>();
            faveImages = new List<string>();

            foreach (string path in lstImages.Items) {
                mainImages.Add(path);
            }
            foreach (string path in lstFavourites.Items) {
                faveImages.Add(path);
            }

            TileSize newTileSizeMode = TileSize.Averaged;
            if (preserveHorizontalToolStripMenuItem.Checked)
                newTileSizeMode = TileSize.PreserveHorizontal;
            else if (preserveVerticalToolStripMenuItem.Checked)
                newTileSizeMode = TileSize.PreserveVertical;

            imagesUnchanged = (oldMain != null) && oldMain.SequenceEqual(mainImages) && oldFave.SequenceEqual(faveImages) && (tileSizeMode == newTileSizeMode);

            tileSizeMode = newTileSizeMode;

            if (blackToolStripMenuItem.Checked)
                Collage.BackColor = Color.Black;
            else if (whiteToolStripMenuItem.Checked)
                Collage.BackColor = Color.White;
            else
                Collage.BackColor = Color.Transparent;

            if (small640x480ToolStripMenuItem.Checked)
                Collage.Size = new SizeF(640, 480);
            else if (medium1024x768ToolStripMenuItem.Checked)
                Collage.Size = new SizeF(1024, 768);
            else
                Collage.Size = new SizeF(1600, 1200);

            Collage.InitialArrangement = randomToolStripMenuItem.Checked ? InitialArrangement.Random : InitialArrangement.Uniform;

            Collage.SoftEdges = softEdgesToolStripMenuItem.Checked;

            UseWaitCursor = true;
            backgroundWorker.RunWorkerAsync();
        }

        void clearImagesToolStripMenuItem_Click(object sender, EventArgs e) {
            lstFavourites.Items.Clear();
            lstImages.Items.Clear();
        }

        void mutualExclusionHandler(object sender, EventArgs e) {
            ToolStripMenuItem item = (ToolStripMenuItem)sender;
            foreach (ToolStripMenuItem other in ((ToolStripMenuItem)item.OwnerItem).DropDownItems.OfType<ToolStripMenuItem>()) {
                other.Checked = (item == other);
            }
        }

        void aboutToolStripMenuItem_Click(object sender, EventArgs e) {
            MessageBox.Show(Properties.Resources.AboutText, "About", MessageBoxButtons.OK, MessageBoxIcon.Information);
        }

        void exitToolStripMenuItem_Click(object sender, EventArgs e) {
            Close();
        }

        void lstFavourites_KeyDown(object sender, KeyEventArgs e) {
            if (e.KeyCode == Keys.Delete) {
                e.Handled = true;
                e.SuppressKeyPress = true;

                foreach (object item in lstFavourites.SelectedItems.OfType<object>().ToList()) {
                    lstFavourites.Items.Remove(item);
                }
            }
        }

        void btnPreview_Click(object sender, EventArgs e) {
            Go(true);
        }

		void backgroundWorker_ProgressChanged(object sender, ProgressChangedEventArgs e) {
			statusLabel.Text = Convert.ToString(e.UserState);
			statusBar.Value = e.ProgressPercentage;
		}

		void backgroundWorker_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e) {
			UseWaitCursor = false;
			statusLabel.Text = "Ready";
            numImagesLabel.Text = String.Format("{0} images", mainImages.Count + faveImages.Count);
			statusBar.Value = 0;

			if (e.Error != null) {
				MessageBox.Show(e.Error.Message, e.Error.GetType().Name, MessageBoxButtons.OK, MessageBoxIcon.Error);
			}
			else {
				picResult.Image = Bitmap;
			}
		}

		void backgroundWorker_DoWork(object sender, DoWorkEventArgs e) {
			backgroundWorker.ReportProgress(25, "Reading image information...");

            if (!imagesUnchanged) {
                Collage.Images.Clear();
                foreach (string path in mainImages) {
                    Collage.Images.Add(new ImageInfo(path, 100f, tileSizeMode));
                }
                foreach (string path in faveImages) {
                    Collage.Images.Add(new ImageInfo(path, 200f, tileSizeMode));
                }
            }

			backgroundWorker.ReportProgress(50, "Calculating layout...");

			if (!skipLayout) Collage.Arrange();

			backgroundWorker.ReportProgress(75, "Rendering collage image...");

            if (previewOnly)
                Bitmap = Collage.Preview();
            else
    			Bitmap = Collage.Render();

			backgroundWorker.ReportProgress(100, "Finished.");
		}

		void lstImages_MouseMove(object sender, MouseEventArgs e) {
			if (e.Button == System.Windows.Forms.MouseButtons.Left) {
				string[] items = lstImages.SelectedItems.OfType<string>().ToArray();
				DataObject o = new DataObject(DataFormats.FileDrop, items);
				lstImages.DoDragDrop(o, DragDropEffects.Move);
			}
		}

		void lstFavourites_MouseMove(object sender, MouseEventArgs e) {
			if (e.Button == System.Windows.Forms.MouseButtons.Left) {
				string[] items = lstFavourites.SelectedItems.OfType<string>().ToArray();
				DataObject o = new DataObject(DataFormats.FileDrop, items);
				lstImages.DoDragDrop(o, DragDropEffects.Move);
			}
		}

		void lstFavourites_Format(object sender, ListControlConvertEventArgs e) {
			e.Value = Path.GetFileName(Convert.ToString(e.ListItem));
		}

		void lstImages_Format(object sender, ListControlConvertEventArgs e) {
			e.Value = Path.GetFileName(Convert.ToString(e.ListItem));
		}

		void lstImages_DragDrop(object sender, DragEventArgs e) {
			if (e.Data.GetDataPresent(DataFormats.FileDrop)) {
				string[] items = (string[])e.Data.GetData(DataFormats.FileDrop);
				foreach (string item in items) {
					lstFavourites.Items.Remove(item);
					lstImages.Items.Add(item);
				}
			}
		}

		void lstImages_DragEnter(object sender, DragEventArgs e) {
			if (e.Data.GetDataPresent(DataFormats.FileDrop)) {
				e.Effect = DragDropEffects.Move;
			}
		}

		void lstFavourites_DragDrop(object sender, DragEventArgs e) {
			if (e.Data.GetDataPresent(DataFormats.FileDrop)) {
				string[] items = (string[])e.Data.GetData(DataFormats.FileDrop);
				foreach (string item in items) {
					lstImages.Items.Remove(item);
					lstFavourites.Items.Add(item);
				}
			}
		}

		void lstFavourites_DragEnter(object sender, DragEventArgs e) {
			if (e.Data.GetDataPresent(DataFormats.FileDrop)) {
				e.Effect = DragDropEffects.Move;
			}
		}

		void lstImages_KeyDown(object sender, KeyEventArgs e) {
			if (e.KeyCode == Keys.Delete) {
				e.Handled = true;
				e.SuppressKeyPress = true;

				foreach (object item in lstImages.SelectedItems.OfType<object>().ToList()) {
					lstImages.Items.Remove(item);
				}
			}
		}

		void btnSave_Click(object sender, EventArgs e) {
			if (Bitmap != null) {
				if (saveFileDialog.ShowDialog() == System.Windows.Forms.DialogResult.OK) {
					Bitmap.Save(saveFileDialog.FileName, System.Drawing.Imaging.ImageFormat.Png);
				}
			}
			else {
				MessageBox.Show("You cannot save until you have generated a collage.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
			}
		}

		void btnOpen_Click(object sender, EventArgs e) {
			if (openFileDialog.ShowDialog() == System.Windows.Forms.DialogResult.OK) {
				lstImages.Items.AddRange(openFileDialog.FileNames);
			}
		}

		void btnGo_Click(object sender, EventArgs e) {
            Go(false);
		}
	}
}
